import { PublicKey } from "@solana/web3.js";
import { encoding, keccak256, toChainId, } from "@wormhole-foundation/sdk-connect";
import { Ntt } from "@wormhole-foundation/sdk-definitions-ntt";
import BN from "bn.js";
export const BPF_LOADER_UPGRADEABLE_PROGRAM_ID = new PublicKey("BPFLoaderUpgradeab1e11111111111111111111111");
export function programDataAddress(programId) {
    return PublicKey.findProgramAddressSync([new PublicKey(programId).toBytes()], BPF_LOADER_UPGRADEABLE_PROGRAM_ID)[0];
}
export const pubKeyConversion = {
    to: (encoded) => new PublicKey(encoded),
    from: (decoded) => decoded.toBytes(),
};
//neither anchor nor solana web3 have a built-in way to parse this, because ofc they don't
export const programDataLayout = [
    { name: "slot", binary: "uint", endianness: "little", size: 8 },
    {
        name: "upgradeAuthority",
        binary: "switch",
        idSize: 1,
        idTag: "isSome",
        layouts: [
            [[0, false], []],
            [
                [1, true],
                [
                    {
                        name: "value",
                        binary: "bytes",
                        size: 32,
                        custom: pubKeyConversion,
                    },
                ],
            ],
        ],
    },
];
export const U64 = {
    MAX: new BN((2n ** 64n - 1n).toString()),
    to: (amount, unit) => {
        const ret = new BN(Math.round(amount * unit));
        if (ret.isNeg())
            throw new Error("Value negative");
        if (ret.bitLength() > 64)
            throw new Error("Value too large");
        return ret;
    },
    from: (amount, unit) => amount.toNumber() / unit,
};
export function derivePda(seeds, programId) {
    const toBytes = (s) => typeof s === "string" ? encoding.bytes.encode(s) : s;
    return PublicKey.findProgramAddressSync(Array.isArray(seeds) ? seeds.map(toBytes) : [toBytes(seeds)], new PublicKey(programId))[0];
}
const chainToBytes = (chain) => encoding.bignum.toBytes(toChainId(chain), 2);
// TODO: memoize?
export const nttAddresses = (programId) => {
    const configAccount = () => derivePda("config", programId);
    const emitterAccount = () => derivePda("emitter", programId);
    const inboxRateLimitAccount = (chain) => derivePda(["inbox_rate_limit", chainToBytes(chain)], programId);
    const inboxItemAccount = (chain, nttMessage) => derivePda(["inbox_item", Ntt.messageDigest(chain, nttMessage)], programId);
    const outboxRateLimitAccount = () => derivePda("outbox_rate_limit", programId);
    const tokenAuthority = () => derivePda("token_authority", programId);
    const peerAccount = (chain) => derivePda(["peer", chainToBytes(chain)], programId);
    const transceiverPeerAccount = (chain) => derivePda(["transceiver_peer", chainToBytes(chain)], programId);
    const registeredTransceiver = (transceiver) => derivePda(["registered_transceiver", transceiver.toBytes()], programId);
    const transceiverMessageAccount = (chain, id) => derivePda(["transceiver_message", chainToBytes(chain), id], programId);
    const wormholeMessageAccount = (outboxItem) => derivePda(["message", outboxItem.toBytes()], programId);
    const sessionAuthority = (sender, args) => derivePda([
        "session_authority",
        sender.toBytes(),
        keccak256(encoding.bytes.concat(encoding.bignum.toBytes(args.amount, 8), chainToBytes(args.recipient.chain), args.recipient.address.toUniversalAddress().toUint8Array(), new Uint8Array([args.shouldQueue ? 1 : 0]))),
    ], programId);
    return {
        configAccount,
        outboxRateLimitAccount,
        inboxRateLimitAccount,
        inboxItemAccount,
        sessionAuthority,
        tokenAuthority,
        emitterAccount,
        wormholeMessageAccount,
        peerAccount,
        transceiverPeerAccount,
        transceiverMessageAccount,
        registeredTransceiver,
    };
};
//# sourceMappingURL=utils.js.map